# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Load, update and manage skills on this device."""
import os
from os.path import basename
from threading import Thread, Event, Lock
from time import monotonic

from ovos_bus_client.apis.enclosure import EnclosureAPI
from ovos_bus_client.client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_config.config import Configuration
from ovos_config.locations import get_xdg_config_save_path
from ovos_plugin_manager.skills import find_skill_plugins
from ovos_plugin_manager.skills import get_skill_directories
from ovos_utils.file_utils import FileWatcher
from ovos_utils.gui import is_gui_connected
from ovos_utils.log import LOG, deprecated
from ovos_utils.network_utils import is_connected_http
from ovos_utils.process_utils import ProcessStatus, StatusCallbackMap, ProcessState
from ovos_workshop.skill_launcher import SKILL_MAIN_MODULE
from ovos_workshop.skill_launcher import SkillLoader, PluginSkillLoader
import warnings


def _shutdown_skill(instance):
    """Shutdown a skill.

    Call the default_shutdown method of the skill, will produce a warning if
    the shutdown process takes longer than 1 second.

    Args:
        instance (MycroftSkill): Skill instance to shutdown
    """
    try:
        ref_time = monotonic()
        # Perform the shutdown
        instance.default_shutdown()

        shutdown_time = monotonic() - ref_time
        if shutdown_time > 1:
            LOG.warning(f'{instance.skill_id} shutdown took {shutdown_time} seconds')
    except Exception:
        LOG.exception(f'Failed to shut down skill: {instance.skill_id}')


def on_started():
    LOG.info('Skills Manager is starting up.')


def on_alive():
    LOG.info('Skills Manager is alive.')


def on_ready():
    LOG.info('Skills Manager is ready.')


def on_error(e='Unknown'):
    LOG.info(f'Skills Manager failed to launch ({e})')


def on_stopping():
    LOG.info('Skills Manager is shutting down...')


class SkillManager(Thread):
    """Manages the loading, activation, and deactivation of Mycroft skills."""

    def __init__(self, bus, watchdog=None, alive_hook=on_alive, started_hook=on_started, ready_hook=on_ready,
                 error_hook=on_error, stopping_hook=on_stopping):
        """Constructor

        Args:
            bus (event emitter): Mycroft messagebus connection
            watchdog (callable): optional watchdog function
            alive_hook (callable): callback function for skill alive status
            started_hook (callable): callback function for skill started status
            ready_hook (callable): callback function for skill ready status
            error_hook (callable): callback function for skill error status
            stopping_hook (callable): callback function for skill stopping status
        """
        super(SkillManager, self).__init__()
        self.bus = bus
        self._settings_watchdog = None
        # Set watchdog to argument or function returning None
        self._watchdog = watchdog or (lambda: None)
        callbacks = StatusCallbackMap(on_started=started_hook,
                                      on_alive=alive_hook,
                                      on_ready=ready_hook,
                                      on_error=error_hook,
                                      on_stopping=stopping_hook)
        self.status = ProcessStatus('skills', callback_map=callbacks)
        self.status.set_started()

        self._lock = Lock()
        self._setup_event = Event()
        self._stop_event = Event()
        self._connected_event = Event()
        self._network_event = Event()
        self._gui_event = Event()
        self._network_loaded = Event()
        self._internet_loaded = Event()
        self._network_skill_timeout = 300
        self._allow_state_reloads = True
        self._logged_skill_warnings = list()
        self._detected_installed_skills = bool(find_skill_plugins())
        if not self._detected_installed_skills:
            LOG.warning(
                "No installed skills detected! if you are running skills in standalone mode ignore this warning,"
                " otherwise you probably want to install skills first!")

        self.config = Configuration()

        self.skill_loaders = {}
        self.plugin_skills = {}
        self.enclosure = EnclosureAPI(bus)
        self.num_install_retries = 0
        self.empty_skill_dirs = set()  # Save a record of empty skill dirs.

        self._define_message_bus_events()
        self.daemon = True

        self.status.bind(self.bus)
        self._init_filewatcher()

    @property
    def blacklist(self):
        """Get the list of blacklisted skills from the configuration.

        Returns:
            list: List of blacklisted skill ids.
        """
        return Configuration().get("skills", {}).get("blacklisted_skills",
                                                     ["skill-ovos-stop.openvoiceos"])

    def _init_filewatcher(self):
        """Initialize the file watcher to monitor skill settings files for changes."""
        sspath = f"{get_xdg_config_save_path()}/skills/"
        os.makedirs(sspath, exist_ok=True)
        self._settings_watchdog = FileWatcher([sspath],
                                              callback=self._handle_settings_file_change,
                                              recursive=True,
                                              ignore_creation=True)

    def _handle_settings_file_change(self, path: str):
        """Handle changes to skill settings files.

        Args:
            path (str): Path to the settings file that has changed.
        """
        if path.endswith("/settings.json"):
            skill_id = path.split("/")[-2]
            LOG.info(f"skill settings.json change detected for {skill_id}")
            self.bus.emit(Message("ovos.skills.settings_changed",
                                  {"skill_id": skill_id}))

    def _sync_skill_loading_state(self):
        """Synchronize the loading state of skills with the current system state."""
        resp = self.bus.wait_for_response(Message("ovos.PHAL.internet_check"))
        network = False
        internet = False
        if not self._gui_event.is_set() and is_gui_connected(self.bus):
            self._gui_event.set()

        if resp:
            if resp.data.get('internet_connected'):
                network = internet = True
            elif resp.data.get('network_connected'):
                network = True
        else:
            LOG.debug("ovos-phal-plugin-connectivity-events not detected, performing direct network checks")
            network = internet = is_connected_http()

        if internet and not self._connected_event.is_set():
            LOG.debug("Notify internet connected")
            self.bus.emit(Message("mycroft.internet.connected"))
        elif network and not self._network_event.is_set():
            LOG.debug("Notify network connected")
            self.bus.emit(Message("mycroft.network.connected"))

    def _define_message_bus_events(self):
        """Define message bus events with handlers defined in this class."""
        # Update upon request
        self.bus.on('skillmanager.list', self.send_skill_list)
        self.bus.on('skillmanager.deactivate', self.deactivate_skill)
        self.bus.on('skillmanager.keep', self.deactivate_except)
        self.bus.on('skillmanager.activate', self.activate_skill)

        # Load skills waiting for connectivity
        self.bus.on("mycroft.network.connected", self.handle_network_connected)
        self.bus.on("mycroft.internet.connected", self.handle_internet_connected)
        self.bus.on("mycroft.gui.available", self.handle_gui_connected)
        self.bus.on("mycroft.network.disconnected", self.handle_network_disconnected)
        self.bus.on("mycroft.internet.disconnected", self.handle_internet_disconnected)
        self.bus.on("mycroft.gui.unavailable", self.handle_gui_disconnected)

    @property
    def skills_config(self):
        """Get the skills service configuration.

        Returns:
            dict: Skills configuration.
        """
        return self.config['skills']

    def handle_gui_connected(self, message):
        """Handle GUI connection event.

        Args:
            message: Message containing information about the GUI connection.
        """
        # Some GUI extensions, such as mobile, may request that skills never unload
        self._allow_state_reloads = not message.data.get("permanent", False)
        if not self._gui_event.is_set():
            LOG.debug("GUI Connected")
            self._gui_event.set()
            self._load_new_skills()

    def handle_gui_disconnected(self, message):
        """Handle GUI disconnection event.

        Args:
            message: Message containing information about the GUI disconnection.
        """
        if self._allow_state_reloads:
            self._gui_event.clear()
            self._unload_on_gui_disconnect()

    def handle_internet_disconnected(self, message):
        """Handle internet disconnection event.

        Args:
            message: Message containing information about the internet disconnection.
        """
        if self._allow_state_reloads:
            self._connected_event.clear()
            self._unload_on_internet_disconnect()

    def handle_network_disconnected(self, message):
        """Handle network disconnection event.

        Args:
            message: Message containing information about the network disconnection.
        """
        if self._allow_state_reloads:
            self._network_event.clear()
            self._unload_on_network_disconnect()

    def handle_internet_connected(self, message):
        """Handle internet connection event.

        Args:
            message: Message containing information about the internet connection.
        """
        if not self._connected_event.is_set():
            LOG.debug("Internet Connected")
            self._network_event.set()
            self._connected_event.set()
            self._load_on_internet()

    def handle_network_connected(self, message):
        """Handle network connection event.

        Args:
            message: Message containing information about the network connection.
        """
        if not self._network_event.is_set():
            LOG.debug("Network Connected")
            self._network_event.set()
            self._load_on_network()

    def load_plugin_skills(self, network=None, internet=None):
        """Load plugin skills based on network and internet status.

        Args:
            network (bool): Network connection status.
            internet (bool): Internet connection status.
        """
        loaded_new = False
        if network is None:
            network = self._network_event.is_set()
        if internet is None:
            internet = self._connected_event.is_set()
        plugins = find_skill_plugins()
        loaded_skill_ids = [basename(p) for p in self.skill_loaders]
        for skill_id, plug in plugins.items():
            if skill_id in self.blacklist:
                if skill_id not in self._logged_skill_warnings:
                    self._logged_skill_warnings.append(skill_id)
                    LOG.warning(f"{skill_id} is blacklisted, it will NOT be loaded")
                    LOG.info(f"Consider uninstalling {skill_id} instead of blacklisting it")
                continue
            if skill_id not in self.plugin_skills and skill_id not in loaded_skill_ids:
                skill_loader = self._get_plugin_skill_loader(skill_id, init_bus=False,
                                                             skill_class=plug)
                requirements = skill_loader.runtime_requirements
                if not network and requirements.network_before_load:
                    continue
                if not internet and requirements.internet_before_load:
                    continue
                self._load_plugin_skill(skill_id, plug)
                loaded_new = True
        return loaded_new

    def _get_internal_skill_bus(self):
        """Get a dedicated skill bus connection per skill.

        Returns:
            MessageBusClient: Internal skill bus.
        """
        if not self.config["websocket"].get("shared_connection", True):
            # See BusBricker skill to understand why this matters.
            # Any skill can manipulate the bus from other skills.
            # This patch ensures each skill gets its own connection that can't be manipulated by others.
            # https://github.com/EvilJarbas/BusBrickerSkill
            bus = MessageBusClient(cache=True)
            bus.run_in_thread()
        else:
            bus = self.bus
        return bus

    def _get_plugin_skill_loader(self, skill_id, init_bus=True, skill_class=None):
        """Get a plugin skill loader.

        Args:
            skill_id (str): ID of the skill.
            init_bus (bool): Whether to initialize the internal skill bus.

        Returns:
            PluginSkillLoader: Plugin skill loader instance.
        """
        bus = None
        if init_bus:
            bus = self._get_internal_skill_bus()
        loader = PluginSkillLoader(bus, skill_id)
        if skill_class:
            loader.skill_class = skill_class
        return loader

    def _load_plugin_skill(self, skill_id, skill_plugin):
        """Load a plugin skill.

        Args:
            skill_id (str): ID of the skill.
            skill_plugin: Plugin skill instance.

        Returns:
            PluginSkillLoader: Loaded plugin skill loader instance if successful, None otherwise.
        """
        skill_loader = self._get_plugin_skill_loader(skill_id, skill_class=skill_plugin)
        try:
            load_status = skill_loader.load(skill_plugin)
        except Exception:
            LOG.exception(f'Load of skill {skill_id} failed!')
            load_status = False
        finally:
            self.plugin_skills[skill_id] = skill_loader

        return skill_loader if load_status else None

    def run(self):
        """Run the skill manager thread."""
        self.status.set_alive()

        self._load_on_startup()

        # trigger a sync so we dont need to wait for the plugin to volunteer info
        self._sync_skill_loading_state()

        if not all((self._network_loaded.is_set(),
                    self._internet_loaded.is_set())):
            self.bus.emit(Message(
                'mycroft.skills.error',
                {'internet_loaded': self._internet_loaded.is_set(),
                 'network_loaded': self._network_loaded.is_set()}))

        self.bus.emit(Message('mycroft.skills.initialized'))

        self.status.set_ready()

        LOG.info("ovos-core is ready! additional skills can now be loaded")

        # Scan the file folder that contains Skills.  If a Skill is updated,
        # unload the existing version from memory and reload from the disk.
        while not self._stop_event.wait(30):
            try:
                self._unload_removed_skills()
                self._load_new_skills()
                self._watchdog()
            except Exception:
                LOG.exception('Something really unexpected has occurred '
                              'and the skill manager loop safety harness was '
                              'hit.')

    def _load_on_network(self):
        """Load skills that require a network connection."""
        if self._detected_installed_skills:  # ensure we have skills installed
            LOG.info('Loading skills that require network...')
            self._load_new_skills(network=True, internet=False)
        self._network_loaded.set()

    def _load_on_internet(self):
        """Load skills that require both internet and network connections."""
        if self._detected_installed_skills:  # ensure we have skills installed
            LOG.info('Loading skills that require internet (and network)...')
            self._load_new_skills(network=True, internet=True)
        self._internet_loaded.set()
        self._network_loaded.set()

    def _unload_on_network_disconnect(self):
        """Unload skills that require a network connection to work."""
        with self._lock:
            for skill_dir in self._get_skill_directories():
                skill_id = os.path.basename(skill_dir)
                skill_loader = self._get_skill_loader(skill_dir, init_bus=False)
                requirements = skill_loader.runtime_requirements
                if requirements.requires_network and \
                        not requirements.no_network_fallback:
                    # Unload skills until the network is back
                    self._unload_skill(skill_dir)

    def _unload_on_internet_disconnect(self):
        """Unload skills that require an internet connection to work."""
        with self._lock:
            for skill_dir in self._get_skill_directories():
                skill_id = os.path.basename(skill_dir)
                skill_loader = self._get_skill_loader(skill_dir, init_bus=False)
                requirements = skill_loader.runtime_requirements
                if requirements.requires_internet and \
                        not requirements.no_internet_fallback:
                    # Unload skills until the internet is back
                    self._unload_skill(skill_dir)

    def _unload_on_gui_disconnect(self):
        """Unload skills that require a GUI to work."""
        with self._lock:
            for skill_dir in self._get_skill_directories():
                skill_id = os.path.basename(skill_dir)
                skill_loader = self._get_skill_loader(skill_dir, init_bus=False)
                requirements = skill_loader.runtime_requirements
                if requirements.requires_gui and \
                        not requirements.no_gui_fallback:
                    # Unload skills until the GUI is back
                    self._unload_skill(skill_dir)

    def _load_on_startup(self):
        """Handle offline skills load on startup."""
        if self._detected_installed_skills:  # ensure we have skills installed
            LOG.info('Loading offline skills...')
            self._load_new_skills(network=False, internet=False)

    def _load_new_skills(self, network=None, internet=None, gui=None):
        """Handle loading of skills installed since startup.

        Args:
            network (bool): Network connection status.
            internet (bool): Internet connection status.
            gui (bool): GUI connection status.
        """
        if network is None:
            network = self._network_event.is_set()
        if internet is None:
            internet = self._connected_event.is_set()
        if gui is None:
            gui = self._gui_event.is_set() or is_gui_connected(self.bus)

        # A lock is used because this can be called via state events or as part of the main loop.
        # There is a possible race condition where this handler would be executing several times otherwise.
        with self._lock:

            loaded_new = self.load_plugin_skills(network=network, internet=internet)

            for skill_dir in self._get_skill_directories():
                replaced_skills = []
                skill_id = os.path.basename(skill_dir)
                skill_loader = self._get_skill_loader(skill_dir, init_bus=False)
                requirements = skill_loader.runtime_requirements
                if not network and requirements.network_before_load:
                    continue
                if not internet and requirements.internet_before_load:
                    continue
                if not gui and requirements.gui_before_load:
                    # TODO - companion PR adding this one
                    continue

                # A local source install is replacing this plugin, unload it!
                if skill_id in self.plugin_skills:
                    LOG.info(f"{skill_id} plugin will be replaced by a local version: {skill_dir}")
                    self._unload_plugin_skill(skill_id)

                for old_skill_dir, skill_loader in self.skill_loaders.items():
                    if old_skill_dir != skill_dir and \
                            skill_loader.skill_id == skill_id:
                        # A higher priority equivalent has been detected!
                        replaced_skills.append(old_skill_dir)

                for old_skill_dir in replaced_skills:
                    # Unload the old skill
                    self._unload_skill(old_skill_dir)

                if skill_dir not in self.skill_loaders:
                    self._load_skill(skill_dir)
                    loaded_new = True

        if loaded_new:
            LOG.info("Requesting padatious intent training")
            try:
                response = self.bus.wait_for_response(Message("mycroft.skills.train"),
                                                      "mycroft.skills.trained",
                                                      timeout=60)  # 60 second timeout
                if not response:
                    LOG.error("Padatious training timed out")
                elif response.data.get('error'):
                    LOG.error(f"Padatious training failed: {response.data['error']}")
            except Exception as e:
                LOG.exception(f"Error during padatious training: {e}")

    def _get_skill_loader(self, skill_directory, init_bus=True):
        """Get a skill loader instance.

        Args:
            skill_directory (str): Directory path of the skill.
            init_bus (bool): Whether to initialize the internal skill bus.

        Returns:
            SkillLoader: Skill loader instance.
        """
        bus = None
        if init_bus:
            bus = self._get_internal_skill_bus()
        return SkillLoader(bus, skill_directory)

    def _load_skill(self, skill_directory):
        """Load an old-style skill.

        Args:
            skill_directory (str): Directory path of the skill.

        Returns:
            SkillLoader: Loaded skill loader instance if successful, None otherwise.
        """
        LOG.warning(f"Found deprecated skill directory: {skill_directory}\n"
                    f"please create a setup.py for this skill")
        skill_id = basename(skill_directory)
        if skill_id in self.blacklist:
            if skill_id not in self._logged_skill_warnings:
                self._logged_skill_warnings.append(skill_id)
                LOG.warning(f"{skill_id} is blacklisted, it will NOT be loaded")
                LOG.info(f"Consider deleting {skill_directory} instead of blacklisting it")
            return None

        skill_loader = self._get_skill_loader(skill_directory)
        try:
            load_status = skill_loader.load()
        except Exception:
            LOG.exception(f'Load of skill {skill_directory} failed!')
            load_status = False
        finally:
            self.skill_loaders[skill_directory] = skill_loader
        if load_status:
            LOG.info(f"Loaded old style skill: {skill_directory}")
        else:
            LOG.error(f"Failed to load old style skill: {skill_directory}")
        return skill_loader if load_status else None

    def _unload_skill(self, skill_dir):
        """Unload a skill.

        Args:
            skill_dir (str): Directory path of the skill.
        """
        if skill_dir in self.skill_loaders:
            skill = self.skill_loaders[skill_dir]
            LOG.info(f'Removing {skill.skill_id}')
            try:
                skill.unload()
            except Exception:
                LOG.exception('Failed to shutdown skill ' + skill.id)
            del self.skill_loaders[skill_dir]

    def _get_skill_directories(self):
        """Get valid skill directories.

        Returns:
            list: List of valid skill directories.
        """
        skillmap = {}
        valid_skill_roots = ["/opt/mycroft/skills"] + get_skill_directories()
        for skills_dir in valid_skill_roots:
            if not os.path.isdir(skills_dir):
                continue
            for skill_id in os.listdir(skills_dir):
                skill = os.path.join(skills_dir, skill_id)
                # NOTE: empty folders mean the skill should NOT be loaded
                if os.path.isdir(skill):
                    skillmap[skill_id] = skill

        for skill_id, skill_dir in skillmap.items():
            # TODO: all python packages must have __init__.py!  Better way?
            # check if folder is a skill (must have __init__.py)
            if SKILL_MAIN_MODULE in os.listdir(skill_dir):
                if skill_dir in self.empty_skill_dirs:
                    self.empty_skill_dirs.discard(skill_dir)
            else:
                if skill_dir not in self.empty_skill_dirs:
                    self.empty_skill_dirs.add(skill_dir)
                    LOG.debug('Found skills directory with no skill: ' +
                              skill_dir)

        return skillmap.values()

    def _unload_removed_skills(self):
        """Shutdown removed skills.

        Finds and unloads skills that were removed from the disk.
        """
        skill_dirs = self._get_skill_directories()
        # Find loaded skills that don't exist on disk
        removed_skills = [
            s for s in self.skill_loaders.keys() if s not in skill_dirs
        ]
        for skill_dir in removed_skills:
            self._unload_skill(skill_dir)
        return removed_skills

    def _unload_plugin_skill(self, skill_id):
        """Unload a plugin skill.

        Args:
            skill_id (str): Identifier of the plugin skill to unload.
        """
        if skill_id in self.plugin_skills:
            LOG.info('Unloading plugin skill: ' + skill_id)
            skill_loader = self.plugin_skills[skill_id]
            if skill_loader.instance is not None:
                try:
                    skill_loader.instance.default_shutdown()
                except Exception:
                    LOG.exception('Failed to shutdown plugin skill: ' + skill_loader.skill_id)
            self.plugin_skills.pop(skill_id)

    def is_alive(self, message=None):
        """Respond to is_alive status request."""
        return self.status.state >= ProcessState.ALIVE

    def is_all_loaded(self, message=None):
        """ Respond to all_loaded status request."""
        return self.status.state == ProcessState.READY

    def send_skill_list(self, message=None):
        """Send list of loaded skills."""
        try:
            message_data = {}
            # TODO handle external skills, OVOSAbstractApp/Hivemind skills are not accounted for
            skills = {**self.skill_loaders, **self.plugin_skills}

            for skill_loader in skills.values():
                message_data[skill_loader.skill_id] = {
                    "active": skill_loader.active and skill_loader.loaded,
                    "id": skill_loader.skill_id}

            self.bus.emit(Message('mycroft.skills.list', data=message_data))
        except Exception:
            LOG.exception('Failed to send skill list')

    def deactivate_skill(self, message):
        """Deactivate a skill."""
        try:
            # TODO handle external skills, OVOSAbstractApp/Hivemind skills are not accounted for
            skills = {**self.skill_loaders, **self.plugin_skills}
            for skill_loader in skills.values():
                if message.data['skill'] == skill_loader.skill_id:
                    LOG.info("Deactivating skill: " + skill_loader.skill_id)
                    skill_loader.deactivate()
                    self.bus.emit(message.response())
        except Exception as err:
            LOG.exception('Failed to deactivate ' + message.data['skill'])
            self.bus.emit(message.response({'error': f'failed: {err}'}))

    def deactivate_except(self, message):
        """Deactivate all skills except the provided."""
        try:
            skill_to_keep = message.data['skill']
            LOG.info(f'Deactivating all skills except {skill_to_keep}')
            # TODO handle external skills, OVOSAbstractApp/Hivemind skills are not accounted for
            skills = {**self.skill_loaders, **self.plugin_skills}
            for skill in skills.values():
                if skill.skill_id != skill_to_keep:
                    skill.deactivate()
            LOG.info('Couldn\'t find skill ' + message.data['skill'])
        except Exception:
            LOG.exception('An error occurred during skill deactivation!')

    def activate_skill(self, message):
        """Activate a deactivated skill."""
        try:
            # TODO handle external skills, OVOSAbstractApp/Hivemind skills are not accounted for
            skills = {**self.skill_loaders, **self.plugin_skills}
            for skill_loader in skills.values():
                if (message.data['skill'] in ('all', skill_loader.skill_id)
                        and not skill_loader.active):
                    skill_loader.activate()
                    self.bus.emit(message.response())
        except Exception as err:
            LOG.exception(f'Couldn\'t activate skill {message.data["skill"]}')
            self.bus.emit(message.response({'error': f'failed: {err}'}))

    def stop(self):
        """Tell the manager to shutdown."""
        self.status.set_stopping()
        self._stop_event.set()

        # Do a clean shutdown of all skills
        for skill_loader in self.skill_loaders.values():
            if skill_loader.instance is not None:
                _shutdown_skill(skill_loader.instance)

        # Do a clean shutdown of all plugin skills
        for skill_id in list(self.plugin_skills.keys()):
            self._unload_plugin_skill(skill_id)

        if self._settings_watchdog:
            self._settings_watchdog.shutdown()

    ############
    # Deprecated stuff
    @deprecated("priority skills have been deprecated for a long time", "1.0.0")
    def load_priority(self):
        warnings.warn(
            "priority skills have been deprecated",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("mycroft.ready event has moved to finished booting skill", "1.0.0")
    def is_device_ready(self):
        """Check if the device is ready by waiting for various services to start.

        Returns:
            bool: True if the device is ready, False otherwise.
        Raises:
            TimeoutError: If the device is not ready within a specified timeout.
        """
        warnings.warn(
            "mycroft.ready event has moved to finished booting skill",
            DeprecationWarning,
            stacklevel=2,
        )
        return True

    @deprecated("mycroft.ready event has moved to finished booting skill", "1.0.0")
    def handle_check_device_readiness(self, message):
        warnings.warn(
            "mycroft.ready event has moved to finished booting skill",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("mycroft.ready event has moved to finished booting skill", "1.0.0")
    def check_services_ready(self, services):
        """Report if all specified services are ready.

        Args:
            services (iterable): Service names to check.
        Returns:
            bool: True if all specified services are ready, False otherwise.
        """
        warnings.warn(
            "mycroft.ready event has moved to finished booting skill",
            DeprecationWarning,
            stacklevel=2,
        )
        return True



# backwards compat imports
from padacioso.opm import PadaciosoPipeline as PadaciosoService, PadaciosoIntent
from padacioso import IntentContainer as FallbackIntentContainer
from ovos_utils.log import log_deprecation
log_deprecation("adapt service moved to 'padacioso.opm'. this import is deprecated", "1.0.0")

import warnings

warnings.warn(
    "adapt service moved to 'padacioso'",
    DeprecationWarning,
    stacklevel=2,
)


# backwards compat imports
from ocp_pipeline.opm import OCPPipelineMatcher, OCPFeaturizer, OCPPlayerProxy
from ovos_utils.log import log_deprecation
log_deprecation("adapt service moved to 'ovos-ocp-pipeline-plugin'. this import is deprecated", "1.0.0")

import warnings

warnings.warn(
    "adapt service moved to 'ovos-ocp-pipeline-plugin'",
    DeprecationWarning,
    stacklevel=2,
)


import os
import re
from os.path import dirname
from threading import Event
from typing import Optional, List

from langcodes import closest_match

from ovos_bus_client.message import Message
from ovos_bus_client.session import SessionManager
from ovos_config.config import Configuration
from ovos_plugin_manager.templates.pipeline import PipelineMatch, PipelinePlugin
from ovos_utils import flatten_list
from ovos_utils.bracket_expansion import expand_template
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.log import LOG
from ovos_utils.parse import match_one


class StopService(PipelinePlugin):
    """Intent Service thats handles stopping skills."""

    def __init__(self, bus):
        self.bus = bus
        self._voc_cache = {}
        self.load_resource_files()
        super().__init__(config=Configuration().get("skills", {}).get("stop") or {})

    def load_resource_files(self):
        base = f"{dirname(__file__)}/locale"
        for lang in os.listdir(base):
            lang2 = standardize_lang_tag(lang)
            self._voc_cache[lang2] = {}
            for f in os.listdir(f"{base}/{lang}"):
                with open(f"{base}/{lang}/{f}", encoding="utf-8") as fi:
                    lines = [expand_template(l) for l in fi.read().split("\n")
                             if l.strip() and not l.startswith("#")]
                    n = f.split(".", 1)[0]
                    self._voc_cache[lang2][n] = flatten_list(lines)

    @staticmethod
    def get_active_skills(message: Optional[Message] = None) -> List[str]:
        """Active skill ids ordered by converse priority
        this represents the order in which stop will be called

        Returns:
            active_skills (list): ordered list of skill_ids
        """
        session = SessionManager.get(message)
        return [skill[0] for skill in session.active_skills]

    def _collect_stop_skills(self, message: Message) -> List[str]:
        """
        Collect skills that can be stopped based on a ping-pong mechanism.
        
        This method determines which active skills can handle a stop request by sending
        a stop ping to each active skill and waiting for their acknowledgment.
        
        Parameters:
            message (Message): The original message triggering the stop request.
        
        Returns:
            List[str]: A list of skill IDs that can be stopped. If no skills explicitly
                      indicate they can stop, returns all active skills.
        
        Notes:
            - Excludes skills that are blacklisted in the current session
            - Uses a non-blocking event mechanism to collect skill responses
            - Waits up to 0.5 seconds for skills to respond
            - Falls back to all active skills if no explicit stop confirmation is received
        """
        sess = SessionManager.get(message)

        want_stop = []
        skill_ids = []

        active_skills = [s for s in self.get_active_skills(message)
                         if s not in sess.blacklisted_skills]

        if not active_skills:
            return want_stop

        event = Event()

        def handle_ack(msg):
            """
            Handle acknowledgment from skills during the stop process.
            
            This method is a nested function used in skill stopping negotiation. It validates and tracks skill responses to a stop request.
            
            Parameters:
                msg (Message): Message containing skill acknowledgment details.
            
            Side Effects:
                - Modifies the `want_stop` list with skills that can handle stopping
                - Updates the `skill_ids` list to track which skills have responded
                - Sets the threading event when all active skills have responded
            
            Notes:
                - Checks if a skill can handle stopping based on multiple conditions
                - Ensures all active skills provide a response before proceeding
            """
            nonlocal event, skill_ids
            skill_id = msg.data["skill_id"]

            # validate the stop pong
            if all((skill_id not in want_stop,
                    msg.data.get("can_handle", True),
                    skill_id in active_skills)):
                want_stop.append(skill_id)

            if skill_id not in skill_ids:  # track which answer we got
                skill_ids.append(skill_id)

            if all(s in skill_ids for s in active_skills):
                # all skills answered the ping!
                event.set()

        self.bus.on("skill.stop.pong", handle_ack)

        # ask skills if they can stop
        for skill_id in active_skills:
            self.bus.emit(message.forward(f"{skill_id}.stop.ping",
                                          {"skill_id": skill_id}))

        # wait for all skills to acknowledge they can stop
        event.wait(timeout=0.5)

        self.bus.remove("skill.stop.pong", handle_ack)
        return want_stop or active_skills

    def stop_skill(self, skill_id: str, message: Message) -> bool:
        """
        Stop a skill's ongoing activities and manage its session state.
        
        Sends a stop command to a specific skill and handles its response, ensuring
        that any active interactions or processes are terminated. The method checks
        for errors, verifies the skill's stopped status, and emits additional signals
        to forcibly abort ongoing actions like conversations, questions, or speech.
        
        Args:
            skill_id (str): Unique identifier of the skill to be stopped.
            message (Message): The original message context containing interaction details.
        
        Returns:
            bool: True if the skill was successfully stopped, False otherwise.
        
        Raises:
            Logs error if skill stop request encounters an issue.
        
        Notes:
            - Emits multiple bus messages to ensure complete skill termination
            - Checks and handles different skill interaction states
            - Supports force-stopping of conversations, questions, and speech
        """
        stop_msg = message.reply(f"{skill_id}.stop")
        result = self.bus.wait_for_response(stop_msg, f"{skill_id}.stop.response")
        if result and 'error' in result.data:
            error_msg = result.data['error']
            LOG.error(f"{skill_id}: {error_msg}")
            return False
        elif result is not None:
            stopped = result.data.get('result', False)
        else:
            stopped = False

        if stopped:
            sess = SessionManager.get(message)
            state = sess.utterance_states.get(skill_id, "intent")
            LOG.debug(f"skill response status: {state}")
            if state == "response":  # TODO this is never happening and it should...
                LOG.debug(f"stopping {skill_id} in middle of get_response!")

            # force-kill any ongoing get_response/converse/TTS - see @killable_event decorator
            self.bus.emit(message.forward("mycroft.skills.abort_question", {"skill_id": skill_id}))
            self.bus.emit(message.forward("ovos.skills.converse.force_timeout", {"skill_id": skill_id}))
            # TODO - track if speech is coming from this skill! not currently tracked
            self.bus.emit(message.reply("mycroft.audio.speech.stop",{"skill_id": skill_id}))

        return stopped

    def match_stop_high(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Handles high-confidence stop requests by matching exact stop vocabulary and managing skill stopping.
        
        Attempts to stop skills when an exact "stop" or "global_stop" command is detected. Performs the following actions:
        - Identifies the closest language match for vocabulary
        - Checks for global stop command when no active skills exist
        - Emits a global stop message if applicable
        - Attempts to stop individual skills if a stop command is detected
        - Disables response mode for stopped skills
        
        Parameters:
            utterances (List[str]): List of user utterances to match against stop vocabulary
            lang (str): Four-letter ISO language code for language-specific matching
            message (Message): Message context for generating appropriate responses
        
        Returns:
            Optional[PipelineMatch]: Match result indicating whether stop was handled, with optional skill and session information
            - Returns None if no stop action could be performed
            - Returns PipelineMatch with handled=True for successful global or skill-specific stop
        
        Raises:
            No explicit exceptions raised, but may log debug/info messages during processing
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None

        sess = SessionManager.get(message)

        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        is_stop = self.voc_match(utterance, 'stop', exact=True, lang=lang)
        is_global_stop = self.voc_match(utterance, 'global_stop', exact=True, lang=lang) or \
                         (is_stop and not len(self.get_active_skills(message)))

        conf = 1.0

        if is_global_stop:
            LOG.info(f"Emitting global stop, {len(self.get_active_skills(message))} active skills")
            # emit a global stop, full stop anything OVOS is doing
            self.bus.emit(message.reply("mycroft.stop", {}))
            return PipelineMatch(handled=True,
                                 match_data={"conf": conf},
                                 skill_id=None,
                                 utterance=utterance)

        if is_stop:
            # check if any skill can stop
            for skill_id in self._collect_stop_skills(message):
                LOG.debug(f"Checking if skill wants to stop: {skill_id}")
                if self.stop_skill(skill_id, message):
                    LOG.info(f"Skill stopped: {skill_id}")
                    sess.disable_response_mode(skill_id)
                    return PipelineMatch(handled=True,
                                         match_data={"conf": conf},
                                         skill_id=skill_id,
                                         utterance=utterance,
                                         updated_session=sess)
        return None

    def match_stop_medium(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Handle stop intent with additional context beyond simple stop commands.
        
        This method processes utterances that contain "stop" or global stop vocabulary but may include
        additional words not explicitly defined in intent files. It performs a medium-confidence
        intent matching for stop requests.
        
        Parameters:
            utterances (List[str]): List of input utterances to analyze
            lang (str): Four-letter ISO language code for localization
            message (Message): Message context for generating appropriate responses
        
        Returns:
            Optional[PipelineMatch]: A pipeline match if the stop intent is successfully processed,
            otherwise None if no stop intent is detected
        
        Notes:
            - Attempts to match stop vocabulary with fuzzy matching
            - Falls back to low-confidence matching if medium-confidence match is inconclusive
            - Handles global stop scenarios when no active skills are present
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None

        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        is_stop = self.voc_match(utterance, 'stop', exact=False, lang=lang)
        if not is_stop:
            is_global_stop = self.voc_match(utterance, 'global_stop', exact=False, lang=lang) or \
                             (is_stop and not len(self.get_active_skills(message)))
            if not is_global_stop:
                return None

        return self.match_stop_low(utterances, lang, message)

    def _get_closest_lang(self, lang: str) -> Optional[str]:
        if self._voc_cache:
            lang = standardize_lang_tag(lang)
            closest, score = closest_match(lang, list(self._voc_cache.keys()))
            # https://langcodes-hickford.readthedocs.io/en/sphinx/index.html#distance-values
            # 0 -> These codes represent the same language, possibly after filling in values and normalizing.
            # 1- 3 -> These codes indicate a minor regional difference.
            # 4 - 10 -> These codes indicate a significant but unproblematic regional difference.
            if score < 10:
                return closest
        return None

    def match_stop_low(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Perform a low-confidence fuzzy match for stop intent before fallback processing.
        
        This method attempts to match stop-related vocabulary with low confidence and handle stopping of active skills.
        
        Parameters:
            utterances (List[str]): List of input utterances to match against stop vocabulary
            lang (str): Four-letter ISO language code for vocabulary matching
            message (Message): Message context used for generating replies and managing session
        
        Returns:
            Optional[PipelineMatch]: A pipeline match object if a stop action is handled, otherwise None
        
        Notes:
            - Increases confidence if active skills are present
            - Attempts to stop individual skills before emitting a global stop signal
            - Handles language-specific vocabulary matching
            - Configurable minimum confidence threshold for stop intent
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None
        sess = SessionManager.get(message)
        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        conf = match_one(utterance, self._voc_cache[lang]['stop'])[1]
        if len(self.get_active_skills(message)) > 0:
            conf += 0.1
        conf = round(min(conf, 1.0), 3)

        if conf < self.config.get("min_conf", 0.5):
            return None

        # check if any skill can stop
        for skill_id in self._collect_stop_skills(message):
            LOG.debug(f"Checking if skill wants to stop: {skill_id}")
            if self.stop_skill(skill_id, message):
                sess.disable_response_mode(skill_id)
                return PipelineMatch(handled=True,
                                     match_data={"conf": conf},
                                     skill_id=skill_id,
                                     utterance=utterance,
                                     updated_session=sess)

        # emit a global stop, full stop anything OVOS is doing
        LOG.debug(f"Emitting global stop signal, {len(self.get_active_skills(message))} active skills")
        self.bus.emit(message.reply("mycroft.stop", {}))
        return PipelineMatch(handled=True,
                             # emit instead of intent message {"conf": conf},
                             match_data={"conf": conf},
                             skill_id=None,
                             utterance=utterance)

    def voc_match(self, utt: str, voc_filename: str, lang: str,
                  exact: bool = False):
        """
        Determine if the given utterance contains the vocabulary provided.

        By default the method checks if the utterance contains the given vocab
        thereby allowing the user to say things like "yes, please" and still
        match against "Yes.voc" containing only "yes". An exact match can be
        requested.

        The method first checks in the current Skill's .voc files and secondly
        in the "res/text" folder of mycroft-core. The result is cached to
        avoid hitting the disk each time the method is called.

        Args:
            utt (str): Utterance to be tested
            voc_filename (str): Name of vocabulary file (e.g. 'yes' for
                                'res/text/en-us/yes.voc')
            lang (str): Language code, defaults to self.lang
            exact (bool): Whether the vocab must exactly match the utterance

        Returns:
            bool: True if the utterance has the given vocabulary it
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return False

        _vocs = self._voc_cache[lang].get(voc_filename) or []

        if utt and _vocs:
            if exact:
                # Check for exact match
                return any(i.strip() == utt
                           for i in _vocs)
            else:
                # Check for matches against complete words
                return any([re.match(r'.*\b' + i + r'\b.*', utt)
                            for i in _vocs])
        return False



from typing import Optional, List
from ovos_config import Configuration
from ovos_plugin_manager.metadata_transformers import find_metadata_transformer_plugins
from ovos_plugin_manager.text_transformers import find_utterance_transformer_plugins

from ovos_utils.json_helper import merge_dict
from ovos_utils.log import LOG


class UtteranceTransformersService:

    def __init__(self, bus, config=None):
        self.config_core = config or Configuration()
        self.loaded_plugins = {}
        self.has_loaded = False
        self.bus = bus
        self.config = self.config_core.get("utterance_transformers") or {}
        self.load_plugins()

    @staticmethod
    def find_plugins():
        return find_utterance_transformer_plugins().items()

    def load_plugins(self):
        for plug_name, plug in self.find_plugins():
            if plug_name in self.config:
                # if disabled skip it
                if not self.config[plug_name].get("active", True):
                    continue
                try:
                    self.loaded_plugins[plug_name] = plug()
                    LOG.info(f"loaded utterance transformer plugin: {plug_name}")
                except Exception as e:
                    LOG.error(e)
                    LOG.exception(f"Failed to load utterance transformer plugin: {plug_name}")

    @property
    def plugins(self):
        """
        Return loaded transformers in priority order, such that modules with a
        higher `priority` rank are called first and changes from lower ranked
        transformers are applied last

        A plugin of `priority` 1 will override any existing context keys and
        will be the last to modify utterances`
        """
        return sorted(self.loaded_plugins.values(),
                      key=lambda k: k.priority, reverse=True)

    def shutdown(self):
        for module in self.plugins:
            try:
                module.shutdown()
            except:
                pass

    def transform(self, utterances: List[str], context: Optional[dict] = None):
        context = context or {}

        for module in self.plugins:
            try:
                utterances, data = module.transform(utterances, context)
                _safe = {k:v for k,v in data.items() if k != "session"}  # no leaking TTS/STT creds in logs    
                LOG.debug(f"{module.name}: {_safe}")
                context = merge_dict(context, data)
            except Exception as e:
                LOG.warning(f"{module.name} transform exception: {e}")
        return utterances, context


class MetadataTransformersService:

    def __init__(self, bus, config=None):
        self.config_core = config or Configuration()
        self.loaded_plugins = {}
        self.has_loaded = False
        self.bus = bus
        self.config = self.config_core.get("metadata_transformers") or {}
        self.load_plugins()

    @staticmethod
    def find_plugins():
        return find_metadata_transformer_plugins().items()

    def load_plugins(self):
        for plug_name, plug in self.find_plugins():
            if plug_name in self.config:
                # if disabled skip it
                if not self.config[plug_name].get("active", True):
                    continue
                try:
                    self.loaded_plugins[plug_name] = plug()
                    LOG.info(f"loaded metadata transformer plugin: {plug_name}")
                except Exception as e:
                    LOG.error(e)
                    LOG.exception(f"Failed to load metadata transformer plugin: {plug_name}")

    @property
    def plugins(self):
        """
        Return loaded transformers in priority order, such that modules with a
        higher `priority` rank are called first and changes from lower ranked
        transformers are applied last.

        A plugin of `priority` 1 will override any existing context keys
        """
        return sorted(self.loaded_plugins.values(),
                      key=lambda k: k.priority, reverse=True)

    def shutdown(self):
        for module in self.plugins:
            try:
                module.shutdown()
            except:
                pass

    def transform(self, context: Optional[dict] = None):
        context = context or {}

        for module in self.plugins:
            try:
                data = module.transform(context)                
                _safe = {k:v for k,v in data.items() if k != "session"}  # no leaking TTS/STT creds in logs    
                LOG.debug(f"{module.name}: {_safe}")
                context = merge_dict(context, data)
            except Exception as e:
                LOG.warning(f"{module.name} transform exception: {e}")
        return context





# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import warnings
from collections import defaultdict
from typing import Tuple, Callable, Union, List

import requests
from ocp_pipeline.opm import OCPPipelineMatcher
from ovos_adapt.opm import AdaptPipeline
from ovos_bus_client.message import Message
from ovos_bus_client.session import SessionManager
from ovos_bus_client.util import get_message_lang
from ovos_config.config import Configuration
from ovos_config.locale import get_valid_languages
from ovos_persona import PersonaService
from ovos_plugin_manager.templates.pipeline import PipelineMatch, IntentHandlerMatch
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.log import LOG, log_deprecation, deprecated
from ovos_utils.metrics import Stopwatch
from ovos_utils.thread_utils import create_daemon
from padacioso.opm import PadaciosoPipeline as PadaciosoService

from ovos_commonqa.opm import CommonQAService
from ovos_core.intent_services.converse_service import ConverseService
from ovos_core.intent_services.fallback_service import FallbackService
from ovos_core.intent_services.stop_service import StopService
from ovos_core.transformers import MetadataTransformersService, UtteranceTransformersService


class IntentService:
    """OVOS intent service. parses utterances using a variety of systems.

    The intent service also provides the internal API for registering and
    querying the intent service.
    """

    def __init__(self, bus, config=None):
        self.bus = bus
        self.config = config or Configuration().get("intents", {})

        # Dictionary for translating a skill id to a name
        self.skill_names = {}

        self._adapt_service = None
        self._padatious_service = None
        self._padacioso_service = None
        self._fallback = None
        self._converse = None
        self._common_qa = None
        self._stop = None
        self._ocp = None
        self._load_pipeline_plugins()

        self.utterance_plugins = UtteranceTransformersService(bus)
        self.metadata_plugins = MetadataTransformersService(bus)

        # connection SessionManager to the bus,
        # this will sync default session across all components
        SessionManager.connect_to_bus(self.bus)

        self.bus.on('recognizer_loop:utterance', self.handle_utterance)

        # Context related handlers
        self.bus.on('add_context', self.handle_add_context)
        self.bus.on('remove_context', self.handle_remove_context)
        self.bus.on('clear_context', self.handle_clear_context)

        # Intents API
        self.registered_vocab = []
        self.bus.on('intent.service.intent.get', self.handle_get_intent)
        self.bus.on('intent.service.skills.get', self.handle_get_skills)
        self.bus.on('mycroft.skills.loaded', self.update_skill_name_dict)

        # internal, track skills that call self.deactivate to avoid reactivating them again
        self._deactivations = defaultdict(list)
        self.bus.on('intent.service.skills.deactivate', self._handle_deactivate)

    def _load_pipeline_plugins(self):
        # TODO - replace with plugin loader from OPM
        self._adapt_service = AdaptPipeline(bus=self.bus, config=self.config.get("adapt", {}))
        if "padatious" not in self.config:
            self.config["padatious"] = Configuration().get("padatious", {})
        try:
            if self.config["padatious"].get("disabled"):
                LOG.info("padatious forcefully disabled in config")
            else:
                from ovos_padatious.opm import PadatiousPipeline
                if "instant_train" not in self.config["padatious"]:
                    self.config["padatious"]["instant_train"] = False
                self._padatious_service = PadatiousPipeline(self.bus, self.config["padatious"])
        except ImportError:
            LOG.error(f'Failed to create padatious intent handlers, padatious not installed')

        # by default only load padacioso is padatious is not available
        # save memory if padacioso isnt needed
        disable_padacioso = self.config.get("disable_padacioso", self._padatious_service is not None)
        if not disable_padacioso:
            self._padacioso_service = PadaciosoService(self.bus, self.config["padatious"])
        elif "disable_padacioso" not in self.config:
            LOG.debug("Padacioso pipeline is disabled, only padatious is loaded. "
                      "set 'disable_padacioso': false in mycroft.conf if you want it to load alongside padatious")
        self._fallback = FallbackService(self.bus)
        self._converse = ConverseService(self.bus)
        self._common_qa = CommonQAService(self.bus, self.config.get("common_query"))
        self._stop = StopService(self.bus)
        self._ocp = OCPPipelineMatcher(self.bus, config=self.config.get("OCP", {}))
        self._persona = PersonaService(self.bus, config=self.config.get("persona", {}))

    def update_skill_name_dict(self, message):
        """Messagebus handler, updates dict of id to skill name conversions."""
        self.skill_names[message.data['id']] = message.data['name']

    def get_skill_name(self, skill_id):
        """Get skill name from skill ID.

        Args:
            skill_id: a skill id as encoded in Intent handlers.

        Returns:
            (str) Skill name or the skill id if the skill wasn't found
        """
        return self.skill_names.get(skill_id, skill_id)

    def _handle_transformers(self, message):
        """
        Pipe utterance through transformer plugins to get more metadata.
        Utterances may be modified by any parser and context overwritten
        """
        lang = get_message_lang(message)  # per query lang or default Configuration lang
        original = utterances = message.data.get('utterances', [])
        message.context["lang"] = lang
        utterances, message.context = self.utterance_plugins.transform(utterances, message.context)
        if original != utterances:
            message.data["utterances"] = utterances
            LOG.debug(f"utterances transformed: {original} -> {utterances}")
        message.context = self.metadata_plugins.transform(message.context)
        return message

    @staticmethod
    def disambiguate_lang(message):
        """ disambiguate language of the query via pre-defined context keys
        1 - stt_lang -> tagged in stt stage  (STT used this lang to transcribe speech)
        2 - request_lang -> tagged in source message (wake word/request volunteered lang info)
        3 - detected_lang -> tagged by transformers  (text classification, free form chat)
        4 - config lang (or from message.data)
        """
        default_lang = get_message_lang(message)
        valid_langs = get_valid_languages()
        valid_langs = [standardize_lang_tag(l) for l in valid_langs]
        lang_keys = ["stt_lang",
                     "request_lang",
                     "detected_lang"]
        for k in lang_keys:
            if k in message.context:
                v = standardize_lang_tag(message.context[k])
                if v in valid_langs:  # TODO - use lang distance instead to choose best dialect
                    if v != default_lang:
                        LOG.info(f"replaced {default_lang} with {k}: {v}")
                    return v
                else:
                    LOG.warning(f"ignoring {k}, {v} is not in enabled languages: {valid_langs}")

        return default_lang

    def get_pipeline(self, skips=None, session=None) -> Tuple[str, Callable]:
        """return a list of matcher functions ordered by priority
        utterances will be sent to each matcher in order until one can handle the utterance
        the list can be configured in mycroft.conf under intents.pipeline,
        in the future plugins will be supported for users to define their own pipeline"""
        session = session or SessionManager.get()

        # Create matchers
        # TODO - from plugins
        padatious_matcher = None
        if self._padatious_service is None:
            needs_pada = any("padatious" in p for p in session.pipeline)
            if self._padacioso_service is not None:
                if needs_pada:
                    LOG.warning("padatious is not available! using padacioso in it's place, "
                                "intent matching will be extremely slow in comparison")
                padatious_matcher = self._padacioso_service
            elif needs_pada:
                LOG.warning("padatious is not available! only adapt (keyword based) intents will match!")
        else:
            padatious_matcher = self._padatious_service

        matchers = {
            "converse": self._converse.converse_with_skills,
            "stop_high": self._stop.match_stop_high,
            "stop_medium": self._stop.match_stop_medium,
            "stop_low": self._stop.match_stop_low,
            "adapt_high": self._adapt_service.match_high,
            "common_qa": self._common_qa.match,
            "fallback_high": self._fallback.high_prio,
            "adapt_medium": self._adapt_service.match_medium,
            "fallback_medium": self._fallback.medium_prio,
            "adapt_low": self._adapt_service.match_low,
            "fallback_low": self._fallback.low_prio,
            "ovos-persona-pipeline-plugin-high": self._persona.match_high,
            "ovos-persona-pipeline-plugin-low": self._persona.match_low
        }
        if self._padacioso_service is not None:
            matchers.update({
                "padacioso_high": self._padacioso_service.match_high,
                "padacioso_medium": self._padacioso_service.match_medium,
                "padacioso_low": self._padacioso_service.match_low,

            })
        if padatious_matcher is not None:
            matchers.update({
                "padatious_high": padatious_matcher.match_high,
                "padatious_medium": padatious_matcher.match_medium,
                "padatious_low": padatious_matcher.match_low,

            })
        if self._ocp is not None:
            matchers.update({
                "ocp_high": self._ocp.match_high,
                "ocp_medium": self._ocp.match_medium,
                "ocp_fallback": self._ocp.match_fallback,
                "ocp_legacy": self._ocp.match_legacy})
        skips = skips or []
        pipeline = [k for k in session.pipeline if k not in skips]
        if any(k not in matchers for k in pipeline):
            LOG.warning(f"Requested some invalid pipeline components! "
                        f"filtered {[k for k in pipeline if k not in matchers]}")
            pipeline = [k for k in pipeline if k in matchers]
        LOG.debug(f"Session pipeline: {pipeline}")
        return [(k, matchers[k]) for k in pipeline]

    @staticmethod
    def _validate_session(message, lang):
        # get session
        lang = standardize_lang_tag(lang)
        sess = SessionManager.get(message)
        if sess.session_id == "default":
            updated = False
            # Default session, check if it needs to be (re)-created
            if sess.expired():
                sess = SessionManager.reset_default_session()
                updated = True
            if lang != sess.lang:
                sess.lang = lang
                updated = True
            if updated:
                SessionManager.update(sess)
                SessionManager.sync(message)
        else:
            sess.lang = lang
            SessionManager.update(sess)
        sess.touch()
        return sess

    def _handle_deactivate(self, message):
        """internal helper, track if a skill asked to be removed from active list during intent match
        in this case we want to avoid reactivating it again
        This only matters in PipelineMatchers, such as fallback and converse
        in those cases the activation is only done AFTER the match, not before unlike intents
        """
        sess = SessionManager.get(message)
        skill_id = message.data.get("skill_id")
        self._deactivations[sess.session_id].append(skill_id)

    def _emit_match_message(self, match: Union[IntentHandlerMatch, PipelineMatch], message: Message, lang: str):
        """
        Emit a reply message for a matched intent, updating session and skill activation.

        This method processes matched intents from either a pipeline matcher or an intent handler,
        creating a reply message with matched intent details and managing skill activation.

        Args:
            match (Union[IntentHandlerMatch, PipelineMatch]): The matched intent object containing
                utterance and matching information.
            message (Message): The original messagebus message that triggered the intent match.
            lang (str): The language of the pipeline plugin match

        Details:
            - Handles two types of matches: PipelineMatch and IntentHandlerMatch
            - Creates a reply message with matched intent data
            - Activates the corresponding skill if not previously deactivated
            - Updates session information
            - Emits the reply message on the messagebus

        Side Effects:
            - Modifies session state
            - Emits a messagebus event
            - Can trigger skill activation events

        Returns:
            None
        """
        reply = None
        try:
            sess = match.updated_session or SessionManager.get(message)
        except AttributeError:  # old ovos-plugin-manager version
            LOG.warning("outdated ovos-plugin-manager detected! please update to version 0.8.0")
            sess = SessionManager.get(message)
        sess.lang = lang  # ensure it is updated

        # utterance fully handled by pipeline matcher
        if isinstance(match, PipelineMatch):
            if match.handled:
                reply = message.reply("ovos.utterance.handled", {"skill_id": match.skill_id})

            # upload intent metrics if enabled
            create_daemon(self._upload_match_data, (match.utterance,
                                                    match.skill_id,
                                                    lang,
                                                    match.match_data))

        # Launch skill if not handled by the match function
        elif isinstance(match, IntentHandlerMatch) and match.match_type:
            # keep all original message.data and update with intent match
            data = dict(message.data)
            data.update(match.match_data)
            reply = message.reply(match.match_type, data)

            # upload intent metrics if enabled
            create_daemon(self._upload_match_data, (match.utterance,
                                                    match.match_type,
                                                    lang,
                                                    match.match_data))

        if reply is not None:
            reply.data["utterance"] = match.utterance
            reply.data["lang"] = lang

            # update active skill list
            if match.skill_id:
                # ensure skill_id is present in message.context
                reply.context["skill_id"] = match.skill_id

                # NOTE: do not re-activate if the skill called self.deactivate
                # we could also skip activation if skill is already active,
                # but we still want to update the timestamp
                was_deactivated = match.skill_id in self._deactivations[sess.session_id]
                if not was_deactivated:
                    sess.activate_skill(match.skill_id)
                    # emit event for skills callback -> self.handle_activate
                    self.bus.emit(reply.forward(f"{match.skill_id}.activate"))

            # update Session if modified by pipeline
            reply.context["session"] = sess.serialize()

            # finally emit reply message
            self.bus.emit(reply)

        else:  # upload intent metrics if enabled
            create_daemon(self._upload_match_data, (match.utterance,
                                                    "complete_intent_failure",
                                                    lang,
                                                    match.match_data))

    @staticmethod
    def _upload_match_data(utterance: str, intent: str, lang: str, match_data: dict):
        """if enabled upload the intent match data to a server, allowing users and developers
        to collect metrics/datasets to improve the pipeline plugins and skills.

        There isn't a default server to upload things too, users needs to explicitly configure one

        https://github.com/OpenVoiceOS/ovos-opendata-server
        """
        config = Configuration().get("open_data", {})
        endpoints: List[str] = config.get("intent_urls", [])  # eg. "http://localhost:8000/intents"
        if not endpoints:
            return  # user didn't configure any endpoints to upload metrics to
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        headers = {"Content-Type": "application/x-www-form-urlencoded",
                   "User-Agent": config.get("user_agent", "ovos-metrics")}
        data = {
            "utterance": utterance,
            "intent": intent,
            "lang": lang,
            "match_data": json.dumps(match_data, ensure_ascii=False)
        }
        for url in endpoints:
            try:
                # Add a timeout to prevent hanging
                response = requests.post(url, data=data, headers=headers, timeout=3)
                LOG.info(f"Uploaded intent metrics to '{url}' - Response: {response.status_code}")
            except Exception as e:
                LOG.warning(f"Failed to upload metrics: {e}")

    def send_cancel_event(self, message):
        """
        Emit events and play a sound when an utterance is canceled.
        
        Logs the cancellation with the specific cancel word, plays a predefined cancel sound,
        and emits multiple events to signal the utterance cancellation.
        
        Parameters:
            message (Message): The original message that triggered the cancellation.
        
        Events Emitted:
            - 'mycroft.audio.play_sound': Plays a cancel sound from configuration
            - 'ovos.utterance.cancelled': Signals that the utterance was canceled
            - 'ovos.utterance.handled': Indicates the utterance processing is complete
        
        Notes:
            - Uses the default cancel sound path 'snd/cancel.mp3' if not specified in configuration
            - Ensures events are sent as replies to the original message
        """
        LOG.info("utterance canceled, cancel_word:" + message.context.get("cancel_word"))
        # play dedicated cancel sound
        sound = Configuration().get('sounds', {}).get('cancel', "snd/cancel.mp3")
        # NOTE: message.reply to ensure correct message destination
        self.bus.emit(message.reply('mycroft.audio.play_sound', {"uri": sound}))
        self.bus.emit(message.reply("ovos.utterance.cancelled"))
        self.bus.emit(message.reply("ovos.utterance.handled"))

    def handle_utterance(self, message: Message):
        """Main entrypoint for handling user utterances

        Monitor the messagebus for 'recognizer_loop:utterance', typically
        generated by a spoken interaction but potentially also from a CLI
        or other method of injecting a 'user utterance' into the system.

        Utterances then work through this sequence to be handled:
        1) UtteranceTransformers can modify the utterance and metadata in message.context
        2) MetadataTransformers can modify the metadata in message.context
        3) Language is extracted from message
        4) Active skills attempt to handle using converse()
        5) Padatious high match intents (conf > 0.95)
        6) Adapt intent handlers
        7) CommonQuery Skills
        8) High Priority Fallbacks
        9) Padatious near match intents (conf > 0.8)
        10) General Fallbacks
        11) Padatious loose match intents (conf > 0.5)
        12) Catch all fallbacks including Unknown intent handler

        If all these fail the complete_intent_failure message will be sent
        and a generic error sound played.

        Args:
            message (Message): The messagebus data
        """
        # Get utterance utterance_plugins additional context
        message = self._handle_transformers(message)

        if message.context.get("canceled"):
            self.send_cancel_event(message)
            return

        # tag language of this utterance
        lang = self.disambiguate_lang(message)

        utterances = message.data.get('utterances', [])
        LOG.info(f"Parsing utterance: {utterances}")

        stopwatch = Stopwatch()

        # get session
        sess = self._validate_session(message, lang)
        message.context["session"] = sess.serialize()

        # match
        match = None
        with stopwatch:
            self._deactivations[sess.session_id] = []
            # Loop through the matching functions until a match is found.
            for pipeline, match_func in self.get_pipeline(session=sess):
                langs = [lang]
                if self.config.get("multilingual_matching"):
                    # if multilingual matching is enabled, attempt to match all user languages if main fails
                    langs += [l for l in get_valid_languages() if l != lang]
                for intent_lang in langs:
                    match = match_func(utterances, intent_lang, message)
                    if match:
                        LOG.info(f"{pipeline} match ({intent_lang}): {match}")
                        if match.skill_id and match.skill_id in sess.blacklisted_skills:
                            LOG.debug(
                                f"ignoring match, skill_id '{match.skill_id}' blacklisted by Session '{sess.session_id}'")
                            continue
                        if isinstance(match, IntentHandlerMatch) and match.match_type in sess.blacklisted_intents:
                            LOG.debug(
                                f"ignoring match, intent '{match.match_type}' blacklisted by Session '{sess.session_id}'")
                            continue
                        try:
                            self._emit_match_message(match, message, intent_lang)
                            break
                        except:
                            LOG.exception(f"{match_func} returned an invalid match")
                else:
                    LOG.debug(f"no match from {match_func}")
                    continue
                break
            else:
                # Nothing was able to handle the intent
                # Ask politely for forgiveness for failing in this vital task
                self.send_complete_intent_failure(message)

        LOG.debug(f"intent matching took: {stopwatch.time}")

        # sync any changes made to the default session, eg by ConverseService
        if sess.session_id == "default":
            SessionManager.sync(message)
        elif sess.session_id in self._deactivations:
            self._deactivations.pop(sess.session_id)
        return match, message.context, stopwatch

    def send_complete_intent_failure(self, message):
        """Send a message that no skill could handle the utterance.

        Args:
            message (Message): original message to forward from
        """
        sound = Configuration().get('sounds', {}).get('error', "snd/error.mp3")
        # NOTE: message.reply to ensure correct message destination
        self.bus.emit(message.reply('mycroft.audio.play_sound', {"uri": sound}))
        self.bus.emit(message.reply('complete_intent_failure'))
        self.bus.emit(message.reply("ovos.utterance.handled"))

    @staticmethod
    def handle_add_context(message: Message):
        """Add context

        Args:
            message: data contains the 'context' item to add
                     optionally can include 'word' to be injected as
                     an alias for the context item.
        """
        entity = {'confidence': 1.0}
        context = message.data.get('context')
        word = message.data.get('word') or ''
        origin = message.data.get('origin') or ''
        # if not a string type try creating a string from it
        if not isinstance(word, str):
            word = str(word)
        entity['data'] = [(word, context)]
        entity['match'] = word
        entity['key'] = word
        entity['origin'] = origin
        sess = SessionManager.get(message)
        sess.context.inject_context(entity)

    @staticmethod
    def handle_remove_context(message: Message):
        """Remove specific context

        Args:
            message: data contains the 'context' item to remove
        """
        context = message.data.get('context')
        if context:
            sess = SessionManager.get(message)
            sess.context.remove_context(context)

    @staticmethod
    def handle_clear_context(message: Message):
        """Clears all keywords from context """
        sess = SessionManager.get(message)
        sess.context.clear_context()

    def handle_get_intent(self, message):
        """Get intent from either adapt or padatious.

        Args:
            message (Message): message containing utterance
        """
        utterance = message.data["utterance"]
        lang = get_message_lang(message)
        sess = SessionManager.get(message)

        # Loop through the matching functions until a match is found.
        for pipeline, match_func in self.get_pipeline(skips=["converse",
                                                             "common_qa",
                                                             "fallback_high",
                                                             "fallback_medium",
                                                             "fallback_low"],
                                                      session=sess):
            match = match_func([utterance], lang, message)
            if match:
                if match.match_type:
                    intent_data = match.match_data
                    intent_data["intent_name"] = match.match_type
                    intent_data["intent_service"] = pipeline
                    intent_data["skill_id"] = match.skill_id
                    intent_data["handler"] = match_func.__name__
                    self.bus.emit(message.reply("intent.service.intent.reply",
                                                {"intent": intent_data}))
                    return

        # signal intent failure
        self.bus.emit(message.reply("intent.service.intent.reply",
                                    {"intent": None}))

    def handle_get_skills(self, message):
        """Send registered skills to caller.

        Argument:
            message: query message to reply to.
        """
        self.bus.emit(message.reply("intent.service.skills.reply",
                                    {"skills": self.skill_names}))

    def shutdown(self):
        self.utterance_plugins.shutdown()
        self.metadata_plugins.shutdown()
        self._adapt_service.shutdown()
        if self._padacioso_service:
            self._padacioso_service.shutdown()
        if self._padatious_service:
            self._padatious_service.shutdown()
        self._common_qa.shutdown()
        self._converse.shutdown()
        self._fallback.shutdown()
        if self._ocp:
            self._ocp.shutdown()

        self.bus.remove('recognizer_loop:utterance', self.handle_utterance)
        self.bus.remove('add_context', self.handle_add_context)
        self.bus.remove('remove_context', self.handle_remove_context)
        self.bus.remove('clear_context', self.handle_clear_context)
        self.bus.remove('mycroft.skills.loaded', self.update_skill_name_dict)
        self.bus.remove('intent.service.intent.get', self.handle_get_intent)
        self.bus.remove('intent.service.skills.get', self.handle_get_skills)

    ###########
    # DEPRECATED STUFF
    @property
    def registered_intents(self):
        log_deprecation("direct access to self.adapt_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        warnings.warn(
            "direct access to self.adapt_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        lang = get_message_lang()
        return [parser.__dict__
                for parser in self._adapt_service.engines[lang].intent_parsers]

    @property
    def adapt_service(self):
        warnings.warn(
            "direct access to self.adapt_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.adapt_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._adapt_service

    @property
    def padatious_service(self):
        warnings.warn(
            "direct access to self.padatious_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.padatious_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._padatious_service

    @property
    def padacioso_service(self):
        warnings.warn(
            "direct access to self.padatious_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.padacioso_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._padacioso_service

    @property
    def fallback(self):
        warnings.warn(
            "direct access to self.fallback is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.fallback is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._fallback

    @property
    def converse(self):
        warnings.warn(
            "direct access to self.converse is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.converse is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._converse

    @property
    def common_qa(self):
        warnings.warn(
            "direct access to self.common_qa is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.common_qa is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._common_qa

    @property
    def stop(self):
        warnings.warn(
            "direct access to self.stop is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.stop is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._stop

    @property
    def ocp(self):
        warnings.warn(
            "direct access to self.ocp is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.ocp is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        return self._ocp

    @adapt_service.setter
    def adapt_service(self, value):
        warnings.warn(
            "direct access to self.adapt_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.adapt_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._adapt_service = value

    @padatious_service.setter
    def padatious_service(self, value):
        warnings.warn(
            "direct access to self.padatious_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.padatious_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._padatious_service = value

    @padacioso_service.setter
    def padacioso_service(self, value):
        warnings.warn(
            "direct access to self.padacioso_service is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.padacioso_service is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._padacioso_service = value

    @fallback.setter
    def fallback(self, value):
        warnings.warn(
            "direct access to self.fallback is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.fallback is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._fallback = value

    @converse.setter
    def converse(self, value):
        warnings.warn(
            "direct access to self.converse is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.converse is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._converse = value

    @common_qa.setter
    def common_qa(self, value):
        warnings.warn(
            "direct access to self.common_qa is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.common_qa is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._common_qa = value

    @stop.setter
    def stop(self, value):
        warnings.warn(
            "direct access to self.stop is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.stop is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._stop = value

    @ocp.setter
    def ocp(self, value):
        warnings.warn(
            "direct access to self.ocp is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        log_deprecation("direct access to self.ocp is deprecated, "
                        "pipelines are in the progress of being replaced with plugins", "1.0.0")
        self._ocp = value

    @deprecated("handle_get_adapt moved to adapt service, this method does nothing", "1.0.0")
    def handle_get_adapt(self, message: Message):
        warnings.warn(
            "moved to adapt service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_adapt_manifest moved to adapt service, this method does nothing", "1.0.0")
    def handle_adapt_manifest(self, message):
        warnings.warn(
            "moved to adapt service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_vocab_manifest moved to adapt service, this method does nothing", "1.0.0")
    def handle_vocab_manifest(self, message):
        warnings.warn(
            "moved to adapt service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_get_padatious moved to padatious service, this method does nothing", "1.0.0")
    def handle_get_padatious(self, message):
        warnings.warn(
            "moved to padatious service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_padatious_manifest moved to padatious service, this method does nothing", "1.0.0")
    def handle_padatious_manifest(self, message):
        warnings.warn(
            "moved to padatious service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_entity_manifest moved to padatious service, this method does nothing", "1.0.0")
    def handle_entity_manifest(self, message):
        warnings.warn(
            "moved to padatious service, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_register_vocab moved to individual pipeline services, this method does nothing", "1.0.0")
    def handle_register_vocab(self, message):
        warnings.warn(
            "moved to pipeline plugins, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_register_intent moved to individual pipeline services, this method does nothing", "1.0.0")
    def handle_register_intent(self, message):
        warnings.warn(
            "moved to pipeline plugins, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_detach_intent moved to individual pipeline services, this method does nothing", "1.0.0")
    def handle_detach_intent(self, message):
        warnings.warn(
            "moved to pipeline plugins, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )

    @deprecated("handle_detach_skill moved to individual pipeline services, this method does nothing", "1.0.0")
    def handle_detach_skill(self, message):
        warnings.warn(
            "moved to pipeline plugins, this method does nothing",
            DeprecationWarning,
            stacklevel=2,
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import os.path

from setuptools import setup, find_packages

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def get_version():
    """ Find the version of ovos-core"""
    version = None
    version_file = os.path.join(BASEDIR, 'ovos_core', 'version.py')
    major, minor, build, alpha = (0, 0, 0, 0)
    with open(version_file) as f:
        for line in f:
            if 'VERSION_MAJOR' in line:
                major = line.split('=')[1].strip()
            elif 'VERSION_MINOR' in line:
                minor = line.split('=')[1].strip()
            elif 'VERSION_BUILD' in line:
                build = line.split('=')[1].strip()
            elif 'VERSION_ALPHA' in line:
                alpha = line.split('=')[1].strip()

            if ((major and minor and build and alpha) or
                    '# END_VERSION_BLOCK' in line):
                break
    version = f"{major}.{minor}.{build}"
    if int(alpha):
        version += f"a{alpha}"
    return version


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    with open(os.path.join(BASEDIR, requirements_file), 'r') as f:
        requirements = f.read().splitlines()
        if 'MYCROFT_LOOSE_REQUIREMENTS' in os.environ:
            print('USING LOOSE REQUIREMENTS!')
            requirements = [r.replace('==', '>=').replace('~=', '>=') for r in requirements]
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]


with open(os.path.join(BASEDIR, "README.md"), "r") as f:
    long_description = f.read()

PLUGIN_ENTRY_POINT = [
    'ovos-converse-pipeline-plugin=ovos_core.intent_services.converse_service:ConverseService',
    'ovos-fallback-pipeline-plugin=ovos_core.intent_services.fallback_service:FallbackService',
    'ovos-stop-pipeline-plugin=ovos_core.intent_services.stop_service:StopService'
]


setup(
    name='ovos-core',
    version=get_version(),
    license='Apache-2.0',
    url='https://github.com/OpenVoiceOS/ovos-core',
    description='The spiritual successor to Mycroft AI, OVOS is flexible voice assistant software that can be run almost anywhere!',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=required('requirements/requirements.txt'),
    extras_require={
        'mycroft': required('requirements/mycroft.txt'),
        'lgpl': required('requirements/lgpl.txt'),
        'plugins': required('requirements/plugins.txt'),
        'skills-essential': required('requirements/skills-essential.txt'),
        'skills-extra': required('requirements/skills-extra.txt'),
        'skills-audio': required('requirements/skills-audio.txt'),
        'skills-desktop': required('requirements/skills-desktop.txt'),
        'skills-internet': required('requirements/skills-internet.txt'),
        'skills-gui': required('requirements/skills-gui.txt'),
        'skills-media': required('requirements/skills-media.txt'),
        'skills-ca': required('requirements/skills-ca.txt'),
        'skills-pt': required('requirements/skills-pt.txt'),
        'skills-en': required('requirements/skills-en.txt')
    },
    packages=find_packages(include=['ovos_core*']),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
    entry_points={
        'opm.pipeline': PLUGIN_ENTRY_POINT,
        'console_scripts': [
            'ovos-core=ovos_core.__main__:main'
        ]
    }
)
